import 'dictionary.dart';
import 'host_function.dart';
import 'stack.dart';
import 'virtual_machine.dart';

typedef DeferredCallback = void Function(VirtualMachine, List<dynamic>);

/// A trick to allow 'infix' functions and similar
/// The deferred execution takes a count of values to accept before
/// executing its callback.
class DeferredExecutionListener extends ValueListener {
  final DeferredCallback callback;

  int seenValueCount;

  DeferredExecutionListener(this.seenValueCount, this.callback);

  final collectedValues = [];

  @override
  invoke(VirtualMachine vm, dynamic value, ValueCallback next) {
    collectedValues.add(value);
    if (collectedValues.length == seenValueCount) {
      vm.removeListener(this);
      callback(vm, collectedValues);
    }
  }
}

/// Like [DeferredExecutionListener], but instead of waiting for a particular
/// number of values to pass it collects them until a specific identifier is seen
class DeferredUntilListener extends ValueListener {
  final String delimiter;
  final DeferredCallback callback;

  DeferredUntilListener(this.delimiter, this.callback);

  final collectedValues = [];

  @override
  dynamic invoke(VirtualMachine vm, dynamic value, ValueCallback next) {
    if (value is Identifier && value.value == delimiter) {
      vm.removeListener(this);
      callback(vm, collectedValues);
      return;
    }
    collectedValues.add(value);
  }
}

typedef InfixDeferredCallback = void Function(
    VirtualMachine vm, dynamic lhs, dynamic rhs);

/// Allow idiomatic execution of vm deferral
extension DeferredExecutionApi on VirtualMachine {
  /// defer a call until we've seen enough values go past
  void defer(int seenValueCount, DeferredCallback callback) {
    pushListener(DeferredExecutionListener(seenValueCount, callback));
  }

  void deferUntil(String delimiter, DeferredCallback callback) {
    pushListener(DeferredUntilListener(delimiter, callback));
  }

  // use a defer to make an infix dart function
  // the lhs and rhs counts are used so that the stack can be rotated so that the
  // left side of the function is popped first (in contrast with a normal pop which would work right to left)
  // this allows more idiomatic stdlib definitions of maths operations
  void infix(DartFunction callback,
      {int rhsValueCount = 1, int lhsValueCount = 1}) {
    defer(rhsValueCount, (vm, values) {
      for (final value in values) {
        vm.eval(value);
      }
      vm.rotate(rhsValueCount + lhsValueCount, lhsValueCount);
      callback(vm);
    });
  }
}
