import 'virtual_machine.dart';

/// A value holding an identifier
class Identifier {
  final String value;

  const Identifier(this.value);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is Identifier &&
          runtimeType == other.runtimeType &&
          value == other.value;

  @override
  int get hashCode => value.hashCode;

  @override
  String toString() {
    return ":$value";
  }
}

/// set an identifier to have a value
class SetLocal {
  final String key;
  final dynamic value;
  final bool shadows;

  SetLocal(this.key, this.value, this.shadows);

  @override
  String toString() => "[$key] ${shadows ? '=' : '-'}> $value";
}

/// if we set a global, we use the bottom dictionary rather than the top
class SetGlobal {
  final String key;
  final dynamic value;

  SetGlobal(this.key, this.value);

  @override
  String toString() => "\$[$key] -> $value";
}

/// Pop the top scope off this dictionary listener
class PopScope {
  const PopScope();

  @override
  String toString() => "pop_scope";
}

/// gets the topmost scope from this dictionary listener
class GetTopScope {
  const GetTopScope();

  @override
  String toString() => "top_scope";
}

class PushNewScope {
  final Dictionary parentScope;

  PushNewScope(this.parentScope);

  @override
  String toString() => "*scope";
}

/// A "stack frame" - or a dictionary of functions, with references to its parent frames so that closures can be implemented
class Dictionary {
  final Dictionary? parent;
  final Map<String, dynamic> data = {};

  Dictionary({this.parent});

  /// Sets a given key-value in this frame.
  /// If shadows is true (default) always sets in this frame
  /// If false, this will find the nearest ancestor that has the key and override it
  /// If no ancestor has the key, it is set locally
  void set(String key, dynamic value, {bool shadows = true}) {
    if (shadows) {
      data[key] = value;
    } else {
      final outerFrame = _frameForKey(key) ?? this;
      outerFrame.set(key, value, shadows: true);
    }
  }

  /// Get the given key from the frame
  dynamic get(String key) => _frameForKey(key)?.data[key];

  /// Gets the nearest frame that contains the given key
  /// Returns null if no frames have the key in them
  Dictionary? _frameForKey(String key) {
    var current = this;

    while (true) {
      if (current.data.keys.contains(key)) {
        return current;
      }
      if (current.parent == null) {
        return null;
      }
      current = current.parent!;
    }
  }
}

/// implements basic word functionality
class DictionaryListener extends ValueListener {
  final scopes = [Dictionary()];

  @override
  invoke(VirtualMachine vm, dynamic value, ValueCallback next) {
    if (value is Identifier) {
      final val = scopes.last.get(value.value);
      if (val == null) {
        throw StateError("Identifier not found: ${value.value}");
      }
      return vm.eval(val);
    }

    if (value is SetLocal) {
      scopes.last.set(value.key, value.value, shadows: value.shadows);
      return;
    }

    if (value is GetTopScope) {
      return scopes.last;
    }

    if (value is PopScope && scopes.length > 1) {
      return scopes.removeLast();
    }

    if (value is PushNewScope) {
      return scopes.add(Dictionary(parent: value.parentScope));
    }

    if (value is SetGlobal) {
      return scopes.first.set(value.key, value.value);
    }

    return next(value);
  }
}

extension DictionaryApi on VirtualMachine {
  void operator []=(String key, dynamic value) => eval(SetGlobal(key, value));

  dynamic operator [](String key) => eval(Identifier(key));

  void set(String key, dynamic value, bool shadows) =>
      eval(SetLocal(key, value, shadows));

  Dictionary scope() => eval(const GetTopScope());

  void popScope() => eval(const PopScope());

  void pushScope(Dictionary parentScope) => eval(PushNewScope(parentScope));
}
