import 'deferred.dart';
import 'dictionary.dart';
import 'virtual_machine.dart';

/// A macro
class Macro {
  final Iterable<dynamic> values;

  Macro(this.values);

  @override
  String toString() => "macro(${values.map((it) => it.toString()).join(" ")})";
}

/// A macro is a simple list of values to be evaluated
/// on its own, it can't do much unless we have a way to define macros.
/// We can define this in the language using a dictionary and a dart function
/// for the start and end macros
class MacroListener extends ValueListener {
  @override
  invoke(VirtualMachine vm, dynamic value, ValueCallback next) {
    if (value is Macro) {
      for (final val in value.values) {
        vm.eval(val);
      }
      return;
    }
    return next(value);
  }
}

/// Okay, we need this for recursive macro/function building
/// it works like a deferedUntil builder, but when we encounter the
/// start of an inner macro/fn, we increase the depth by one.
/// we only actually end the defer when we've seen as many ends as starts
class MacroBuilderListener extends ValueListener {
  final Identifier start;
  final Identifier end;
  final DeferredCallback callback;

  MacroBuilderListener(this.start, this.end, this.callback);

  final collectedValues = [];
  int depth = 1;

  @override
  invoke(VirtualMachine vm, value, ValueCallback next) {
    if (value == start) {
      depth += 1;
    } else if (value == end) {
      depth -= 1;
      if (depth == 0) {
        vm.removeListener(this);
        callback(vm, collectedValues);
        return;
      }
    }
    collectedValues.add(value);
  }
}

extension MacroBuilderApi on VirtualMachine {
  void buildMacro(Identifier start, Identifier end, DeferredCallback callback) {
    pushListener(MacroBuilderListener(start, end, callback));
  }
}
