import '../dictionary.dart';

extension on String {
  Iterable<String?> asCharSequence() sync* {
    for (var i = 0; i < length; i++) {
      yield this[i];
    }
    yield '\n'; // always add a trailing newline
  }
}

const _whitespace = [" ", "\t", "\r", "\n"];

/// an error has occurred while parsing
class ParserError extends Error {
  final int line;
  final int char;
  final String message;

  ParserError(this.line, this.char, this.message);

  @override
  String toString() => "Parser Error (line: $line, char: $char): $message";
}

/// The parser
/// Works in a similar way to Forth's but with native string parsing so you don't need the space
/// If it encounters a string, it is parsed
/// Other input is split on whitespaces, each word is checked for compatibility with double
/// if it parses, double is returned, otherwise it is treated as an identifier
class Parser {
  int _line = 1;
  int _char = 0;

  // Parse the input string
  Iterable<dynamic> parse(String input) =>
      _parse(input.asCharSequence().iterator);

  Iterable<dynamic> _parse(Iterator<String?> iterator) sync* {
    final builder = StringBuffer();

    while (iterator.moveNext()) {
      final char = iterator.current;

      _char++;
      if (_char == "\n") {
        _line++;
      }

      if (_whitespace.contains(char)) {
        if (builder.isNotEmpty) {
          yield _tryParse(builder.toString());
          builder.clear();
        }
        continue;
      } else if (char == "'" || char == '"') {
        if (builder.isNotEmpty) {
          throw ParserError(_line, _char,
              "Start of string not preceeded by whitespace: `${builder.toString()}$char`");
        }
        yield _parseString(iterator, char!);
      } else {
        builder.write(char);
      }
    }
  }

  String _parseString(Iterator<String?> iterator, String delimiter) {
    final builder = StringBuffer();
    while (iterator.moveNext()) {
      if (iterator.current == delimiter) {
        return builder.toString();
      }
      if (iterator.current == '\\') {
        iterator.moveNext();
        builder.write(_stringEscapeMap[iterator.current]);
        continue;
      }
      builder.write(iterator.current);
    }
    throw ParserError(_line, _char, "End of file reached while reading string");
  }

  static const _stringEscapeMap = {
    'a': '\a',
    'b': '\b',
    'e': '\e',
    'f': '\f',
    'n': '\n',
    'r': '\r',
    't': '\t',
    'v': '\v',
    '\\': '\\',
    "'": "'",
    '"': '"',
    "?": "?"
  };

  dynamic _tryParse(String string) {
    final asNumber = double.tryParse(string);
    if (asNumber == null) {
      return Identifier(string);
    }
    return asNumber;
  }
}
