import 'deferred.dart';
import 'dictionary.dart';
import 'function.dart';
import 'macro.dart';
import 'stack.dart';
import 'virtual_machine.dart';

/// set here as the difference is in shadowing only
void _assignVar(VirtualMachine vm, String key, value, bool shadows) {
  if (value is FlightFunction) {
    value.name = key;
  }
  vm.set(key, value, shadows);
}

final Map<String, dynamic> _stdlib = {
  // comments
  "/*": (VirtualMachine vm) => vm.deferUntil("*/", (_vm, _values) {}),
  // Declare a macro
  "<{": (VirtualMachine vm) => vm.buildMacro(
        const Identifier("<{"),
        const Identifier("}>"),
        (vm, values) => vm.push(Macro(values)),
      ),
  // declare a function
  "{": (VirtualMachine vm) => vm.buildMacro(
        const Identifier("{"),
        const Identifier("}"),
        (vm, values) {
          vm.push(FlightFunction(Macro(values), vm.scope()));
        },
      ),
  // execute whatever is on top of the stack (anonymous macro and function calls)
  "()": (VirtualMachine vm) => vm.eval(vm.pop()),
  // assign a var
  "->": (VirtualMachine vm) => vm.defer(
      1, (vm, values) => _assignVar(vm, values.first.value, vm.pop(), false)),
  "=>": (VirtualMachine vm) => vm.defer(
      1, (vm, values) => _assignVar(vm, values.first.value, vm.pop(), true)),
  // stack functions
  // dup can just be a command instance as we evaluate it on load
  "pop": const Pop(),
  "dup": const Duplicate(),
  //arithmetic and comparison operators
  "_+": (VirtualMachine vm) => vm.push(vm.pop() + vm.pop()),
  "+": (VirtualMachine vm) => vm.infix((vm) => vm["_+"]),
  "-": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() - vm.pop())),
  "_*": (VirtualMachine vm) => vm.push(vm.pop() * vm.pop()),
  "*": (VirtualMachine vm) => vm.infix((vm) => vm["_*"]),
  "/": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() / vm.pop())),
  "==": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() == vm.pop())),
  "!=": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() != vm.pop())),
  ">": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() > vm.pop())),
  ">=": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() >= vm.pop())),
  "<": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() < vm.pop())),
  "<=": (VirtualMachine vm) => vm.infix((vm) => vm.push(vm.pop() <= vm.pop())),
  "<<": (VirtualMachine vm) =>
      vm.infix((vm) => vm.push(vm.pop().toString() + vm.pop().toString())),
  "print": (VirtualMachine vm) => print(vm.pop()),
  // conditionals
  "true": true,
  "false": false,
  "if": (VirtualMachine vm) {
    final elseFn = vm.pop();
    final thenFn = vm.pop();
    final predicate = vm.pop();
    vm.eval(predicate ? thenFn : elseFn);
  },
  // min max without an array
  "min_n": (VirtualMachine vm) => vm.defer(1, (vm, values) {
        var count = values.first as double;
        var min = double.maxFinite;
        while (count-- > 0) {
          final value = vm.pop();
          if (value < min) min = value;
        }
        vm.push(min);
      }),
  "max_n": (VirtualMachine vm) => vm.defer(1, (vm, values) {
        var count = values.first as double;
        var min = -double.maxFinite;
        while (count-- > 0) {
          final value = vm.pop();
          if (value > min) min = value;
        }
        vm.push(min);
      }),
};

extension StdLib on VirtualMachine {
  void loadStdLib() {
    if (!hasListener(DictionaryListener)) {
      throw StateError(
          "Cannot Load StdLib without a storage mechanism"); // This may not be ideal, we don't care how storage is implemented as long as it is... Do we need a capabilities interface?
    }
    _stdlib.forEach((key, value) {
      this[key] = value;
    });
  }
}
